#include "stdafx.h"
#include "IniUtil.h"

/* Typedefs */
typedef struct _INIREFITEM {
	BYTE	signature[4];
	LPWSTR	pIniBuf;
} INIREFITEM;

/* Defines */
#define INIREF_SIGNATURE	"INIU"
#define	INIUTIL_DELIM_CR	L'\r'
#define	INIUTIL_DELIM_LF	L'\n'
#define	INIUTIL_DELIM_BRA	L'['
#define	INIUTIL_DELIM_KET	L']'
#define	INIUTIL_DELIM_EQ	L'='

static
LPWSTR
GetOneLine(
	LPWSTR	pFileBuf,	// IN
	LPDWORD	pdwSize		// OUT
)
{
	LPWSTR	pNextPos = NULL;

	pNextPos = wcschr(pFileBuf, INIUTIL_DELIM_CR);
	if (pNextPos != NULL) {

		*pdwSize = pNextPos - pFileBuf;

		pNextPos++;
		if (*pNextPos == INIUTIL_DELIM_LF) {
			pNextPos++;
		}
	}

	return pNextPos;
}


static
BOOL
IsSectionMatch(
	LPWSTR	pLine,
	DWORD	dwSize,
	LPWSTR	pSection
)
{
	BOOL	bRet;
	LPWSTR	p;
	DWORD	dwSectionLen;

	bRet = FALSE;

	dwSectionLen = wcslen(pSection);

	if (dwSize == dwSectionLen + 2) {
		p = pLine;

		if (*p == INIUTIL_DELIM_BRA) {
			p++;

			if (wcsnicmp(p, pSection, dwSectionLen) == 0) {
				p += dwSectionLen;

				if (*p == INIUTIL_DELIM_KET) {
					bRet = TRUE;
				}
			}
		}
	}

	return bRet;
}

static
BOOL
IsSectionLine(
	LPWSTR	pLine,
	DWORD	dwSize
)
{
	BOOL	bRet;
	LPWSTR	p;
	DWORD	i;
	DWORD	dwRestSize;

	bRet = FALSE;

	if (dwSize > 2) {
		p = pLine;
		dwRestSize = dwSize;

		if (*p == INIUTIL_DELIM_BRA) {
			p++;
			dwRestSize--;

			for (i = 0; i < dwRestSize; i++) {
				if (*p == INIUTIL_DELIM_KET) {
					bRet = TRUE;
					break;
				}
				p++;
			}
		}
	}

	return bRet;
}

static
LPWSTR
FindSection(
	LPWSTR	pFileBuf,
	LPWSTR	pSection
)
{
	LPWSTR	pLine		= NULL;
	LPWSTR	pNextLine	= NULL;
	LPWSTR	pRet		= NULL;
	DWORD	dwSize;

	pLine = pFileBuf;

	while (pLine != NULL) {
		pNextLine = GetOneLine(pLine, &dwSize);
		if (IsSectionMatch(pLine, dwSize, pSection)) {
			pRet = pNextLine;
			break;
		}

		pLine = pNextLine;
	}

	return pRet;
}

static
LPWSTR
FindEntry(
	LPWSTR	pFileBuf,
	LPWSTR	pKey,
	LPWSTR*	ppValue,
	LPDWORD	pdwValueLen
)
{
	LPWSTR	pLine		= NULL;
	LPWSTR	pNextLine	= NULL;
	LPWSTR	pRet		= NULL;
	DWORD	dwSize;
	DWORD	dwKeyLen;

	dwKeyLen = wcslen(pKey);
	pLine = pFileBuf;

	while (pLine != NULL) {
		pNextLine = GetOneLine(pLine, &dwSize);
		if (IsSectionLine(pLine, dwSize)) {
			break;
		}
		else {
			if (dwSize > dwKeyLen) {
				LPWSTR	p;
				DWORD	dwRestSize;

				p = pLine;
				dwRestSize = dwSize;

				if(wcsnicmp(p, pKey, dwKeyLen) == 0) {
					p += dwKeyLen;
					dwRestSize -= dwKeyLen;

					if (dwRestSize > 0) {
						if (*p == INIUTIL_DELIM_EQ) {
							p++;
							dwRestSize--;

							*ppValue = p;
							*pdwValueLen = dwRestSize;
							pRet = pNextLine;

							break;
						}
					}
				}

			}
		}

		pLine = pNextLine;
	}

	return pRet;
}

static
BOOL
GetIniString(
	LPWSTR	pAppName,			// IN		Section
	LPWSTR	pKeyName,			// IN		Key
	LPWSTR	pReturnedString,	// OUT		擾
	LPDWORD	pdwSize,			// IN, OUT	obt@̕(NULL܂)
	LPWSTR	pFileBuf			// IN		Init@C̃f[^
)
{
	BOOL	bRet;
	LPWSTR	pBuf		= NULL;
	LPWSTR	pValue		= NULL;
	DWORD	dwValueLen	= 0;

	bRet = FALSE;

	if (pAppName == NULL || pKeyName == NULL || pdwSize == NULL) {
		return FALSE;
	}

	pBuf = FindSection(pFileBuf, pAppName);
	if (pBuf != NULL) {
		if (FindEntry(pBuf, pKeyName, &pValue, &dwValueLen) != NULL) {
			if (pReturnedString != NULL) {
				if (*pdwSize > dwValueLen) {
					memcpy(pReturnedString, pValue, dwValueLen*sizeof(WCHAR));
					pReturnedString[dwValueLen] = L'\0';

					bRet = TRUE;
				}
			}
			else {
				bRet = TRUE;
			}

			*pdwSize = dwValueLen + 1;
		}
	}

	return bRet;
}

static
LPWSTR
LoadIniFileToBuffer(
	LPTSTR	pFileName
)
{
	HANDLE	hFile			= INVALID_HANDLE_VALUE;
	LPSTR	pAnsiFileBuf	= NULL;
	LPWSTR	pFileBuf		= NULL;
	int		iBufSize		= 0;

	hFile = CreateFile(pFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile != INVALID_HANDLE_VALUE) {

		BY_HANDLE_FILE_INFORMATION	fileInfo;

		memset(&fileInfo, 0, sizeof(fileInfo));

		if (GetFileInformationByHandle(hFile, &fileInfo) && fileInfo.nFileSizeLow > 0) {
			pAnsiFileBuf = (LPSTR) malloc(fileInfo.nFileSizeLow);
			if (pAnsiFileBuf != NULL) {
				DWORD	dwReaded = 0;

				if (ReadFile(hFile, pAnsiFileBuf, fileInfo.nFileSizeLow, &dwReaded, NULL)) {
					iBufSize = MultiByteToWideChar(CP_ACP, 0, pAnsiFileBuf, fileInfo.nFileSizeLow, NULL, 0);

					if (iBufSize > 0) {
						pFileBuf = (LPWSTR) malloc((iBufSize + 1)*sizeof(WCHAR));
						if (pFileBuf != NULL) {
							MultiByteToWideChar(CP_ACP, 0, pAnsiFileBuf, fileInfo.nFileSizeLow, pFileBuf, iBufSize + 1);
						}
					}

				}

				free(pAnsiFileBuf);
				pAnsiFileBuf = NULL;
			}
		}

		CloseHandle(hFile);
		hFile = NULL;
	}

	return pFileBuf;
}

static
void
DiscardIniBuffer(
	LPWSTR	pFileBuf
)
{
	if (pFileBuf != NULL) {
		free(pFileBuf);
	}
}

static
BOOL
CheckIniRefItem(
	INIREFITEM*	pRef
)
{
	BOOL	bRet;

	bRet = FALSE;

	if (pRef != NULL) {
		if (memcmp(pRef->signature, INIREF_SIGNATURE, sizeof(pRef->signature)) == 0) {
			bRet =  TRUE;
		}
	}

	return bRet;
}

HANDLE
IniUtilCreateRef()
{
	INIREFITEM*	pRef = NULL;

	pRef = (INIREFITEM*) malloc(sizeof(INIREFITEM));
	if (pRef != NULL) {
		memset(pRef, 0, sizeof(INIREFITEM));
		
		memcpy(pRef->signature, INIREF_SIGNATURE, sizeof(pRef->signature));
		pRef->pIniBuf = NULL;
	}

	return (HANDLE)pRef;
}

BOOL
IniUtilDeleteRef(
	HANDLE hRef
)
{
	BOOL	bRet;

	bRet = FALSE;
	if (hRef != NULL) {
		INIREFITEM*	pRef = (INIREFITEM*)hRef;

		if (CheckIniRefItem(pRef)) {
			if (pRef->pIniBuf != NULL) {
				free(pRef->pIniBuf);
				pRef->pIniBuf = NULL;
			}

			free(pRef);
			bRet = TRUE;
		}
	}

	return bRet;
}

BOOL
IniUtilLoadFile(
	HANDLE	hRef,
	LPTSTR	pFileName
)
{
	BOOL	bRet;

	bRet = FALSE;
	if (hRef != NULL) {
		INIREFITEM*	pRef = (INIREFITEM*)hRef;

		if (CheckIniRefItem(pRef)) {
			LPWSTR	pIniBuf = NULL;

			if (pRef->pIniBuf != NULL) {
				free(pRef->pIniBuf);
				pRef->pIniBuf = NULL;
			}

			pIniBuf = LoadIniFileToBuffer(pFileName);
			if (pIniBuf != NULL) {
				pRef->pIniBuf = pIniBuf;
				bRet = TRUE;
			}
		}
	}

	return bRet;
}

BOOL
IniUtilGetString(
	HANDLE	hRef,
	LPWSTR	pAppName,
	LPWSTR	pKeyName,
	LPWSTR	pReturnedString,
	LPDWORD	pdwSize
)
{
	BOOL	bRet;

	bRet = FALSE;
	if (hRef != NULL) {
		INIREFITEM*	pRef = (INIREFITEM*)hRef;

		if (CheckIniRefItem(pRef)) {
			if (pRef->pIniBuf != NULL) {
				bRet = GetIniString(pAppName, pKeyName, pReturnedString, pdwSize, pRef->pIniBuf);
			}
		}
	}

	return bRet;
}
